# coding: utf-8
#

import json
import os
import platform
import queue
import subprocess
import sys
import io
import base64
import time
import traceback
# `pip install futures` for python2
from concurrent.futures import ThreadPoolExecutor
from subprocess import PIPE
import six
import tornado
from tornado.concurrent import run_on_executor
from logzero import logger

from ..device import connect_device, get_device
from ..utils import tostr
from ..version import __version__


class BaseHandler(tornado.web.RequestHandler):
    def set_default_headers(self):
        self.set_header("Access-Control-Allow-Origin", "*")
        self.set_header("Access-Control-Allow-Headers", "x-requested-with")
        self.set_header("Access-Control-Allow-Credentials",
                        "true")  # allow cookie
        self.set_header('Access-Control-Allow-Methods',
                        'POST, GET, PUT, DELETE, OPTIONS')

    def options(self, *args):
        self.set_status(204)  # no body
        self.finish()


class VersionHandler(BaseHandler):
    def get(self):
        ret = {
            'name': "weditor",
            'version': __version__,
        }
        self.write(ret)


class MainHandler(BaseHandler):
    def get(self):
        self.render("index.html")


gqueue = queue.Queue()


class BuildWSHandler(tornado.websocket.WebSocketHandler):
    executor = ThreadPoolExecutor(max_workers=4)

    # proc = None

    def open(self):
        print("Websocket opened")
        self.proc = None

    def check_origin(self, origin):
        return True

    @run_on_executor
    def _run(self, device_url, code):
        """
        Thanks: https://gist.github.com/mosquito/e638dded87291d313717
        """
        try:

            print("DEBUG: run code\n%s" % code)
            env = os.environ.copy()
            env['UIAUTOMATOR_DEBUG'] = 'true'
            if device_url and device_url != 'default':
                env['ATX_CONNECT_URL'] = tostr(device_url)
            start_time = time.time()

            self.proc = subprocess.Popen([sys.executable, "-u"],
                                         env=env,
                                         stdout=PIPE,
                                         stderr=subprocess.STDOUT,
                                         stdin=PIPE)
            self.proc.stdin.write(code)
            self.proc.stdin.close()

            for line in iter(self.proc.stdout.readline, b''):
                print("recv subprocess:", repr(line))
                if line is None:
                    break
                gqueue.put((self, {"buffer": line.decode('utf-8')}))
            print("Wait exit")
            exit_code = self.proc.wait()
            duration = time.time() - start_time
            ret = {
                "buffer": "",
                "result": {
                    "exitCode": exit_code,
                    "duration": int(duration) * 1000
                }
            }
            gqueue.put((self, ret))
            time.sleep(3)  # wait until write done
        except Exception:
            traceback.print_exc()

    @tornado.gen.coroutine
    def on_message(self, message):
        jdata = json.loads(message)
        if self.proc is None:
            code = jdata['content']
            device_url = jdata.get('deviceUrl')
            yield self._run(device_url, code.encode('utf-8'))
            self.close()
        else:
            self.proc.terminate()
            # on Windows, kill is alais of terminate()
            if platform.system() == 'Windows':
                return
            yield tornado.gen.sleep(0.5)
            if self.proc.poll():
                return
            yield tornado.gen.sleep(1.2)
            if self.proc.poll():
                return
            print("Force to kill")
            self.proc.kill()

    def on_close(self):
        print("Websocket closed")


class DeviceConnectHandler(BaseHandler):
    def post(self):
        platform = self.get_argument("platform").lower()
        device_url = self.get_argument("deviceUrl")

        try:
            id = connect_device(platform, device_url)
        except RuntimeError as e:
            self.set_status(410)  # 410 Gone
            self.write({
                "success": False,
                "description": str(e),
            })
        except Exception as e:
            logger.warning("device connect error: %s", e)
            self.set_status(410)  # 410 Gone
            self.write({
                "success": False,
                "description": traceback.format_exc(),
            })
        else:
            ret = {
                "deviceId": id,
                'success': True,
            }
            if platform == "android":
                ws_addr = get_device(id).device.address.replace("http://", "ws://") # yapf: disable
                ret['screenWebSocketUrl'] = ws_addr + "/minicap"
            self.write(ret)


class DeviceHierarchyHandler(BaseHandler):
    def get(self, device_id):
        d = get_device(device_id)
        self.write(d.dump_hierarchy())


class DeviceScreenshotHandler(BaseHandler):
    def get(self, serial):
        logger.info("Serial: %s", serial)
        try:
            d = get_device(serial)
            buffer = io.BytesIO()
            d.screenshot().convert("RGB").save(buffer, format='JPEG')
            b64data = base64.b64encode(buffer.getvalue())
            response = {
                "type": "jpeg",
                "encoding": "base64",
                "data": b64data.decode('utf-8'),
            }
            self.write(response)
        except EnvironmentError as e:
            traceback.print_exc()
            self.set_status(430, "Environment Error")
            self.write({"description": str(e)})
        except RuntimeError as e:
            self.set_status(410)  # Gone
            self.write({"description": traceback.print_exc()})


class StringBuffer():
    def __init__(self):
        self.encoding = 'utf-8'
        self.buf = io.BytesIO()

    def write(self, data):
        if isinstance(data, six.string_types):
            data = data.encode(self.encoding)
        self.buf.write(data)

    def getvalue(self):
        return self.buf.getvalue().decode(self.encoding)


class DeviceCodeDebugHandler(BaseHandler):
    _global = {}

    def run(self, device, code):
        buffer = StringBuffer()
        sys.stdout = buffer
        sys.stderr = buffer

        try:
            is_eval = True
            compiled_code = None
            try:
                compiled_code = compile(code, "<string>", "eval")
            except SyntaxError:
                is_eval = False
                compiled_code = compile(code, "<string>", "exec")

            self._global.update(d=device, time=time, os=os)
            if is_eval:
                ret = eval(code, self._global)
                buffer.write((">>> " + repr(ret) + "\n"))
            else:
                exec(compiled_code, self._global)
        except Exception:
            buffer.write(traceback.format_exc())
        finally:
            sys.stdout = sys.__stdout__
            sys.stderr = sys.__stderr__

        return buffer.getvalue()

    def post(self, device_id):
        d = get_device(device_id)
        code = self.get_argument('code')

        start = time.time()
        output = self.run(d.device, code)

        self.write({
            "success": True,
            "duration": int((time.time() - start) * 1000),
            "content": output,
        })



#! /usr/bin/env python
# -*- encoding: utf-8 -*-

from __future__ import absolute_import, print_function

import argparse
import base64
import hashlib
import io
import json
import os
import platform
import queue
import signal
import socket
import subprocess
import sys
import time
import traceback
import uuid
import webbrowser
# `pip install futures` for python2
from concurrent.futures import ThreadPoolExecutor
from subprocess import PIPE

import requests
import six
import tornado.escape
import tornado.ioloop
import tornado.web
import tornado.websocket
from logzero import logger
from tornado import gen
from tornado.concurrent import run_on_executor
from tornado.escape import json_encode
from tornado.log import enable_pretty_logging

from .web.handlers.page import (BaseHandler, BuildWSHandler,
                                DeviceCodeDebugHandler, DeviceConnectHandler,
                                DeviceHierarchyHandler,
                                DeviceScreenshotHandler, MainHandler,
                                VersionHandler)
from .web.handlers.proxy import StaticProxyHandler
from .web.utils import current_ip, tostr

enable_pretty_logging()

__dir__ = os.path.dirname(os.path.abspath(__file__))

is_closing = False


def signal_handler(signum, frame):
    global is_closing
    print('exiting...')
    is_closing = True


def try_exit():
    global is_closing
    if is_closing:  # clean up here
        tornado.ioloop.IOLoop.instance().stop()
        print('exit success')


def make_app(settings={}):
    application = tornado.web.Application([
        (r"/", MainHandler),
        (r"/api/v1/version", VersionHandler),
        (r"/api/v1/connect", DeviceConnectHandler),
        (r"/api/v1/devices/([^/]+)/screenshot", DeviceScreenshotHandler),
        (r"/api/v1/devices/([^/]+)/hierarchy", DeviceHierarchyHandler),
        (r"/api/v1/devices/([^/]+)/exec", DeviceCodeDebugHandler),
        # cache static assets
        (r"/proxy/https/(.*)", StaticProxyHandler),
        (r"/unpkg.com/.*", StaticProxyHandler),
        (r"/cdn.jsdelivr.net/.*", StaticProxyHandler),
        (r"/ws/v1/build", BuildWSHandler),
    ], **settings)
    return application


def check_running(port: int):
    """
    sys.exit if already running
    """
    try:
        r = requests.get(f"http://localhost:{port}/api/v1/version", timeout=2.0)
        if r.status_code == 200:
            version = r.json().get("version", "dev")
            sys.exit(f"Another weditor({version}) is already running")
    except requests.exceptions.ConnectionError:
        pass
    except Exception as e:
        print("Unknown error: %r" % e)


def run_web(debug=False, port=17310, open_browser=False):
    check_running(port)

    if open_browser:
        # webbrowser.open(url, new=2)
        webbrowser.open(f'http://localhost:{port}', new=2)

    application = make_app({
        'static_path': os.path.join(__dir__, 'static'),
        'template_path': os.path.join(__dir__, 'templates'),
        'debug': debug,
    })
    print('listening on http://%s:%d' % (current_ip(), port))
    if debug:
        logger.info("enable debug mode")
    signal.signal(signal.SIGINT, signal_handler)
    application.listen(port)
    tornado.ioloop.PeriodicCallback(try_exit, 100).start()
    # tornado.ioloop.IOLoop.instance().add_callback(consume_queue)
    tornado.ioloop.IOLoop.instance().start()


def create_shortcut():
    if os.name != 'nt':
        sys.exit("Only valid in Windows")

    import pythoncom  # pyint: disable=import-error
    from win32com.shell import shell
    from win32com.shell import shellcon
    # Refs
    # - https://github.com/pearu/iocbio/blob/master/installer/utils.py
    # - https://blog.csdn.net/thundor/article/details/5968581
    ilist = shell.SHGetSpecialFolderLocation(0, shellcon.CSIDL_DESKTOP)
    dtpath = shell.SHGetPathFromIDList(ilist).decode('utf-8')

    shortcut = pythoncom.CoCreateInstance(shell.CLSID_ShellLink, None,
                                          pythoncom.CLSCTX_INPROC_SERVER,
                                          shell.IID_IShellLink)
    launch_path = sys.executable
    shortcut.SetPath(launch_path)
    shortcut.SetArguments("-m weditor")
    shortcut.SetDescription(launch_path)
    shortcut.SetIconLocation(sys.executable, 0)
    shortcut.QueryInterface(pythoncom.IID_IPersistFile).Save(
        dtpath + "\\WEditor.lnk", 0)
    print("Shortcut created. " + dtpath + "\\WEditor.lnk")


def main():
    ap = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    ap.add_argument('-q',
                    '--quiet',
                    action='store_true',
                    help='quite mode, no open new browser')
    ap.add_argument('-d',
                    '--debug',
                    action='store_true',
                    help='open debug mode')
    ap.add_argument('--shortcut',
                    action='store_true',
                    help='create shortcut in desktop')
    ap.add_argument('-p',
                    '--port',
                    type=int,
                    default=17310,
                    help='local listen port for weditor')

    args = ap.parse_args()
    if args.shortcut:
        create_shortcut()
        return

    open_browser = not args.quiet
    run_web(args.debug, args.port, open_browser)


if __name__ == '__main__':
    main()



# coding: utf-8

import re
import xml.dom.minidom
import uuid

sample_android_page_xml = '''<?xml version="1.0" ?>
<hierarchy rotation="0">
  <node bounds="[0,0][720,1280]" checkable="false" checked="false" class="android.widget.FrameLayout" clickable="false" content-desc="" enabled="true" focusable="false" focused="false" index="0" long-clickable="false" package="com.huawei.android.launcher" password="false" resource-id="" scrollable="false" selected="false" text="">
    <node bounds="[0,0][720,1280]" checkable="false" checked="false" class="android.view.View" clickable="false" content-desc="第 1 屏，共 4 屏" enabled="true" focusable="false" focused="false" index="0" long-clickable="false" package="com.huawei.android.launcher" password="false" resource-id="com.huawei.android.launcher:id/workspace" scrollable="true" selected="false" text="">
      <node bounds="[0,0][720,1110]" checkable="false" checked="false" class="android.view.View" clickable="true" content-desc="" enabled="true" focusable="false" focused="false" index="0" long-clickable="true" package="com.huawei.android.launcher" password="false" resource-id="com.huawei.android.launcher:id/workspace_screen" scrollable="false" selected="false" text="">
        <node bounds="[8,66][184,270]" checkable="false" checked="false" class="android.widget.TextView" clickable="true" content-desc="" enabled="true" focusable="true" focused="false" index="1" long-clickable="true" package="com.huawei.android.launcher" password="false" resource-id="" scrollable="false" selected="false" text="梦幻西游"/>
        <node bounds="[184,66][360,270]" checkable="false" checked="false" class="android.widget.TextView" clickable="true" content-desc="" enabled="true" focusable="true" focused="false" index="2" long-clickable="true" package="com.huawei.android.launcher" password="false" resource-id="" scrollable="false" selected="false" text="梦幻西游"/>
        <node bounds="[360,66][536,270]" checkable="false" checked="false" class="android.widget.TextView" clickable="true" content-desc="" enabled="true" focusable="true" focused="false" index="3" long-clickable="true" package="com.huawei.android.launcher" password="false" resource-id="" scrollable="false" selected="false" text="梦幻西游"/>
      </node>
    </node>
    <node NAF="true" bounds="[0,1083][720,1155]" checkable="false" checked="false" class="android.widget.ImageView" clickable="true" content-desc="" enabled="true" focusable="true" focused="false" index="1" long-clickable="false" package="com.huawei.android.launcher" password="false" resource-id="com.huawei.android.launcher:id/dock_divider" scrollable="false" selected="false" text=""/>
    <node bounds="[0,1110][720,1280]" checkable="false" checked="false" class="android.widget.FrameLayout" clickable="false" content-desc="" enabled="true" focusable="false" focused="false" index="2" long-clickable="false" package="com.huawei.android.launcher" password="false" resource-id="com.huawei.android.launcher:id/hotseat" scrollable="false" selected="false" text="">
      <node bounds="[16,1110][176,1280]" checkable="false" checked="false" class="android.widget.TextView" clickable="true" content-desc="拨号" enabled="true" focusable="true" focused="false" index="0" long-clickable="true" package="com.huawei.android.launcher" password="false" resource-id="" scrollable="false" selected="false" text=""/>
      <node bounds="[192,1110][352,1280]" checkable="false" checked="false" class="android.widget.TextView" clickable="true" content-desc="联系人" enabled="true" focusable="true" focused="false" index="1" long-clickable="true" package="com.huawei.android.launcher" password="false" resource-id="" scrollable="false" selected="false" text=""/>
      <node bounds="[368,1110][528,1280]" checkable="false" checked="false" class="android.widget.TextView" clickable="true" content-desc="信息" enabled="true" focusable="true" focused="false" index="2" long-clickable="true" package="com.huawei.android.launcher" password="false" resource-id="" scrollable="false" selected="false" text=""/>
      <node bounds="[544,1110][704,1280]" checkable="false" checked="false" class="android.widget.TextView" clickable="true" content-desc="浏览器" enabled="true" focusable="true" focused="false" index="3" long-clickable="true" package="com.huawei.android.launcher" password="false" resource-id="" scrollable="false" selected="false" text=""/>
      <node NAF="true" bounds="[0,1110][720,1280]" checkable="false" checked="false" class="android.widget.ImageView" clickable="true" content-desc="" enabled="true" focusable="true" focused="false" index="4" long-clickable="false" package="com.huawei.android.launcher" password="false" resource-id="com.huawei.android.launcher:id/bg_dock" scrollable="false" selected="false" text=""/>
    </node>
  </node>
</hierarchy>
'''


def parse_bounds(text):
    m = re.match(r'\[(\d+),(\d+)\]\[(\d+),(\d+)\]', text)
    if m is None:
        return None
    (lx, ly, rx, ry) = map(int, m.groups())
    return dict(x=lx, y=ly, width=rx - lx, height=ry - ly)


def str2bool(v):
    return v.lower() in ("yes", "true", "t", "1")


def str2int(v):
    return int(v)


def convstr(v):
    return v


__alias = {
    'class': '_type',
    'resource-id': 'resourceId',
    'content-desc': 'description',
    'long-clickable': 'longClickable',
    'bounds': 'rect',
}

__parsers = {
    '_type': convstr, # node className
    # Android
    'rect': parse_bounds,
    'text': convstr,
    'resourceId': convstr,
    'package': convstr,
    'checkable': str2bool,
    'scrollable': str2bool,
    'focused': str2bool,
    'clickable': str2bool,
    'selected': str2bool,
    'longClickable': str2bool,
    'focusable': str2bool,
    'password': str2bool,
    'index': int,
    'description': convstr,
    # iOS
    'name': convstr,
    'label': convstr,
    'x': str2int,
    'y': str2int,
    'width': str2int,
    'height': str2int,
    # iOS && Android
    'enabled': str2bool,
}


def parse_uiautomator_node(node):
    ks = {}
    for key, value in node.attributes.items():
        key = __alias.get(key, key)
        f = __parsers.get(key)
        if value is None:
            ks[key] = None
        elif f:
            ks[key] = f(value)
    if 'bounds' in ks:
        lx, ly, rx, ry = map(int, ks.pop('bounds'))
        ks['rect'] = dict(x=lx, y=ly, width=rx - lx, height=ry - ly)
    return ks


def get_android_hierarchy(d):
    """
    Returns:
        JSON object
    """
    page_xml = d.dump_hierarchy(compressed=False, pretty=False).encode('utf-8')
    dom = xml.dom.minidom.parseString(page_xml)
    root = dom.documentElement

    def travel(node):
        """ return current node info """
        if node.attributes is None:
            return
        json_node = parse_uiautomator_node(node)
        json_node['_id'] = str(uuid.uuid4())
        if node.childNodes:
            children = []
            for n in node.childNodes:
                child = travel(n)
                if child:
                    # child["_parent"] = json_node["_id"]
                    children.append(child)
            json_node['children'] = children
        return json_node

    return travel(root)


def get_ios_hierarchy(d, scale):
    sourcejson = d.source(format='json')

    def travel(node):
        node['_id'] = str(uuid.uuid4())
        node['_type'] = node.pop('type', "null")
        if node.get('rect'):
            rect = node['rect']
            nrect = {}
            for k, v in rect.items():
                nrect[k] = v * scale
            node['rect'] = nrect

        for child in node.get('children', []):
            travel(child)
        return node

    return travel(sourcejson)


def get_webview_hierarchy(d):
    pass


# coding: utf-8
#

import hashlib
import os
import socket

import six


def tostr(s, encoding='utf-8'):
    if six.PY2:
        return s.encode(encoding)
    if isinstance(s, bytes):
        return s.decode(encoding)
    return s


def read_file_content(filename, default=''):
    if not os.path.isfile(filename):
        return default
    with open(filename, 'rb') as f:
        return f.read()


def sha_file(path):
    sha = hashlib.sha1()
    with open(path, 'rb') as f:
        while True:
            data = f.read(65536)
            if not data:
                break
            sha.update(data)
    return sha.hexdigest()


def write_file_content(filename, content):
    with open(filename, 'w') as f:
        f.write(content.encode('utf-8'))


def virt2real(path):
    return os.path.join(os.getcwd(), path.lstrip('/'))


def real2virt(path):
    return os.path.relpath(path, os.getcwd()).replace('\\', '/')


def current_ip():
    with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as s:
        s.connect(("8.8.8.8", 80))
        return s.getsockname()[0]



# coding: utf-8
#

import os
import hashlib
import tornado.httpclient
from logzero import logger

from .page import BaseHandler


class StaticProxyHandler(BaseHandler):
    http_client = tornado.httpclient.AsyncHTTPClient()

    async def get(self, urlpath=None):
        # print(self.request.remote_ip)
        if urlpath is None:
            urlpath = "https://"+self.request.path.lstrip("/")
        else:
            urlpath = "https://" + urlpath

        #logger.debug("path: %s", urlpath)
        m = hashlib.md5()
        m.update(urlpath.encode())

        cache_dir = os.path.expanduser("~/.weditor/cache")
        _, ext = os.path.splitext(urlpath)
        cache_path = os.path.join(cache_dir, m.hexdigest() + ext)
        if not os.path.exists(cache_path):
            request = tornado.httpclient.HTTPRequest(
                url=urlpath,
                method="GET",
                validate_cert=False # fix certificate validate error
            )
            response = await self.http_client.fetch(request)
            if not os.path.isdir(cache_dir):
                os.makedirs(cache_dir)

            with open(cache_path, 'wb') as f:
                f.write(response.body)

        with open(cache_path, 'rb') as f:
            #logger.debug("use cache assets: %s", urlpath)
            self.write(f.read())
